/*
 * mod_zipvfs: http://modzipvfs.sourceforge.net/
 *
 * Apache2 API compliant module, version 0.1 (24-Jul-2002)
 *
 * Copyright (c) 2002 Pavel Kral, All rights reserved.
 * Distributed under GNU GENERAL PUBLIC LICENCE see copying for more
 *
 * This module represent zip compatabile archive (PkWare,WinZip,Jar) as
 * virtual filesystem, allow traverzal throught and read content. This
 * is especially usefull for large documentations (Java JDK). Support for
 * sending raw data if browser has ability to read gziped content and for
 * normal decompressed content. Support for automatics directory listing
 * and index searching.
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_protocol.h"
#include "apr_strings.h"

#include "unzip.h"
#include "iowin32.h"
#include "mod_zipvfs.h"
#include "mod_msgs.h"


zlib_filefunc_def zipFuncs;


typedef struct {
	unsigned char	id1;
	unsigned char	id2;
	unsigned char	method;
	unsigned char	flags;
	unsigned char	modtime[4];
	unsigned char	xfl;
	unsigned char	os;	
} gzipResponseHeader;


apr_status_t zipCleanup(void *cleanedData) {

	if(cleanedData) {
		unzCloseCurrentFile((unzFile)cleanedData);
		unzClose((unzFile)cleanedData);
	}

	return APR_SUCCESS;
}


/*
 *
 *
 *
 */

const char * getContentType(zipResourceRequest *zipRRequest) {	
	request_rec *newRequest;
	
	newRequest = ap_sub_req_lookup_file(zipRRequest->currentFileName,zipRRequest->request,NULL);

	if(newRequest->content_type) 
		return newRequest->content_type; else 
		return HTTP_UNKNOWN_CONTENT;

}

time_t getFileTime(zipResourceRequest *zipRRequest)
{
    struct tm ti = {0,0,0,0,0,0,0,0,0};    

    ti.tm_year = zipRRequest->currentFileInfo.tmu_date.tm_year;
    ti.tm_mon  = zipRRequest->currentFileInfo.tmu_date.tm_mon;
    ti.tm_mday = zipRRequest->currentFileInfo.tmu_date.tm_mday;
    ti.tm_hour = zipRRequest->currentFileInfo.tmu_date.tm_hour,
    ti.tm_min  = zipRRequest->currentFileInfo.tmu_date.tm_min;
    ti.tm_sec  = zipRRequest->currentFileInfo.tmu_date.tm_sec;

    if(ti.tm_year < 80) 
		ti.tm_year += 100; else 
			if(ti.tm_year >= 1900)
				ti.tm_year -= 1900;    

    return mktime(&ti);
}


void listZipDirectory(zipResourceRequest *zipRRequest) {

	unz_global_info zipGlobalInfo;
	unz_file_info fileInfo;
	char *directoryPath;	
	char zipedFilename[256+1];
	char prevFilename[256+1];
	unsigned int dirPathLen;
	unsigned int zipNameLen;
	unsigned int prevNameLen;
	unsigned int suffixPos;
	unsigned int fileCounter;
	unsigned int index;

	// TODO: update time to zip file time attribut
	//
	//ap_update_mtime(zipRRequest->request,
	//ap_set_last_modified(zipRRequest->request);

	zipRRequest->request->content_type = HTTP_HTML_CONTENT;

	if( unzGetGlobalInfo(zipRRequest->zipFile, &zipGlobalInfo) != UNZ_OK ) {
		sendErrorMsg(zipRRequest,"Can't retreive global information about archive.");
		return;
	}

	sendBeginHtmlData(zipRRequest);

	ap_rprintf(zipRRequest->request,
		"<H3>"
		"Directory listing for %s"
		"</H3><P>", zipRRequest->request->path_info);

	// far,far from perfect, but current version of unzLibrary support only all files listing, so
	// there is implemented directory 'filter'
	// no sort is currently implemented

	directoryPath = apr_pstrdup(zipRRequest->request->pool,zipRRequest->request->path_info+1);
	dirPathLen = strlen(directoryPath);

	if( unzGoToFirstFile(zipRRequest->zipFile) != UNZ_OK) {
		ap_rprintf(zipRRequest->request,"Error moving to first file in archive ...");
	} else {

		fileCounter = 0;
		*prevFilename = 0;

		for(index=0; index < zipGlobalInfo.number_entry; index++) {

		if( unzGetCurrentFileInfo(zipRRequest->zipFile,&fileInfo,
						zipedFilename,sizeof(zipedFilename),NULL,0,NULL,0) != UNZ_OK) {
			ap_rprintf(zipRRequest->request,"Error requesting file info ...");		
		}
		

		if( strncmp(directoryPath, zipedFilename, dirPathLen) == 0) {

			if( fileCounter++ || !dirPathLen ) {

				prevNameLen = strlen(prevFilename);

				if(!prevNameLen || strncmp(prevFilename,zipedFilename,prevNameLen) != 0) {
					
					strcpy(prevFilename,zipedFilename);

					zipNameLen = strlen(zipedFilename) - 1;
					suffixPos = zipNameLen;

					if( zipedFilename[suffixPos] == '/' ) suffixPos--;
				    while ( (suffixPos > 0) && (zipedFilename[suffixPos] != '/') ) suffixPos--;
					if( zipedFilename[suffixPos] == '/' ) suffixPos++;

					if( zipedFilename[zipNameLen] == '/' )

						ap_rprintf(zipRRequest->request,"<FONT face=\"Arial\" size=\"2\"><A href=\"%s\">%s</A></FONT><BR>",
							&zipedFilename[suffixPos],zipedFilename); else

						ap_rprintf(zipRRequest->request,"<A href=\"%s\">%s</A><BR>",
						&zipedFilename[suffixPos],zipedFilename);				
				}
			}
			
		}
	
		if (index < zipGlobalInfo.number_entry - 2)
		    if( unzGoToNextFile(zipRRequest->zipFile) != UNZ_OK) {
			ap_rprintf(zipRRequest->request,"Error moving to next file in archive ...");		
			break;
			}

		}
	}

	sendEndHtmlData(zipRRequest);

}

void sendUncompResource(zipResourceRequest *zipRRequest) {
	char *sendBuffer;
	int  readBytes;	
	time_t fileTime;		
	
	sendBuffer = apr_palloc(zipRRequest->request->pool,SEND_BUFFER_SIZE);

	if(!sendBuffer) {
		sendErrorMsg(zipRRequest,"Can't allocate memory resources for uncompress send buffer");
		return;
	}

	if( unzOpenCurrentFile(zipRRequest->zipFile) != UNZ_OK ) {
		sendErrorMsg(zipRRequest,"Can't open current file in archive (normal open)");
		return;
	}
			
	// set HTTP response headers
	//

	zipRRequest->request->content_type = getContentType(zipRRequest);
	ap_set_content_length(zipRRequest->request, zipRRequest->currentFileInfo.uncompressed_size);
	fileTime = getFileTime(zipRRequest);
	ap_update_mtime(zipRRequest->request, fileTime);


	do{
		readBytes = unzReadCurrentFile(zipRRequest->zipFile,sendBuffer,SEND_BUFFER_SIZE);

		if(readBytes > 0)
			ap_rwrite(sendBuffer,readBytes,zipRRequest->request);
	} while(readBytes > 0);

	unzCloseCurrentFile(zipRRequest->zipFile);	


};

void sendCompResource(zipResourceRequest *zipRRequest) {

	char *sendBuffer;
	unsigned int compMethod,compLevel,readBytes,fileStrLen;	
	time_t fileTime;
	gzipResponseHeader gzipHeader =		

#ifdef ENABLE_CRC_SUPPORT
	{0x1F,0x8B,GZIP_DEFLATE_METHOD,GZIP_FLAG_CRC | GZIP_FLAG_FNAME,{0,0,0,0},0,0xFF};
	unsigned long crc;
#else
	{0x1F,0x8B,GZIP_DEFLATE_METHOD, GZIP_FLAG_FNAME,{0,0,0,0},0,0xFF};
#endif
	
	sendBuffer = apr_palloc(zipRRequest->request->pool,SEND_BUFFER_SIZE);

	if(!sendBuffer) {
		sendErrorMsg(zipRRequest,"Can't allocate memory resources for compress send buffer");
		return;
	}

	if( unzOpenCurrentFile2(zipRRequest->zipFile, &compMethod, &compLevel, TRUE) != UNZ_OK ) {
		sendErrorMsg(zipRRequest,"Can't open current file in archive (raw open)");
		return;
	}
		
	// set HTTP response headers
	//

	fileStrLen = strlen(zipRRequest->currentFileName) + 1;

	zipRRequest->request->content_type = getContentType(zipRRequest);
	apr_table_setn(zipRRequest->request->headers_out,HTTP_CONTENT_ENCODING,HTTP_ENCODE_GZIP1);

	fileTime = getFileTime(zipRRequest);
	ap_update_mtime(zipRRequest->request, fileTime);

	// set gzip member format
	//

	// if ASCII bit in zip is present, set FTEXT (optional indication)
	if(zipRRequest->currentFileInfo.internal_fa & ZIP_INTFLAG_ASCII == ZIP_INTFLAG_ASCII)
		gzipHeader.flags |= GZIP_FLAG_FTEXT;

	// zip do gzip compression type transformation
	compLevel = zipRRequest->currentFileInfo.flag & 6;

	if(compLevel == ZIP_NORMAL_COMPRESSION)
		gzipHeader.xfl = XFL_NORMAL_COMPRESSION; else
		if(compLevel == ZIP_MAX_COMPRESSION)
			gzipHeader.xfl = XFL_MAX_COMPRESSION; else
			if((compLevel == ZIP_FAST_COMPRESSION ) || (compLevel == ZIP_SFAST_COMPRESSION))
				gzipHeader.xfl = XFL_FAST_COMPRESSION; else
					gzipHeader.xfl = XFL_NORMAL_COMPRESSION;

    // upper byte indicate the host system for file
	// lover the version number of the software used to encode the file	
	gzipHeader.os = (char)(zipRRequest->currentFileInfo.version >> 8);
	
	// set most recent modification time of compressed file
	gzipHeader.modtime[0] = fileTime & 0xFF;
	gzipHeader.modtime[1] = (fileTime >> 8) & 0xFF;
	gzipHeader.modtime[2] = (fileTime >> 16) & 0xFF;
	gzipHeader.modtime[3] = (fileTime >> 24) & 0xFF;

	// write gzip member header to response
	ap_rwrite(&gzipHeader,sizeof(gzipHeader),zipRRequest->request);
	
	// write FNAME (file name) part of optional member header to response
	ap_rwrite(zipRRequest->currentFileName,fileStrLen,zipRRequest->request);

#ifdef ENABLE_CRC_SUPPORT
	crc = crc32(0L, Z_NULL, 0);
	crc = crc32(crc, (char*)	&gzipHeader, sizeof(gzipHeader));
	crc = crc32(crc, zipRRequest->currentFileName, fileStrLen);


	// write FHCRC (crc16) part of optional member header to response
	ap_rwrite(&crc, 2, zipRRequest->request);
#endif
	
	// write compressed data via sendBuffer

	do{
		readBytes = unzReadCurrentFile(zipRRequest->zipFile,sendBuffer,SEND_BUFFER_SIZE);

		if(readBytes > 0)
			ap_rwrite(sendBuffer,readBytes,zipRRequest->request);
	} while(readBytes > 0);
	

	// write CRC32 (crc32) of uncompressed data
	ap_rwrite(&zipRRequest->currentFileInfo.crc, 4, zipRRequest->request);

	// and finally write ISIZE (input uncompressed size) of file
	ap_rwrite(&zipRRequest->currentFileInfo.uncompressed_size, 4, zipRRequest->request);

	unzCloseCurrentFile(zipRRequest->zipFile);	
};

/*
 * getResourceData
 *
 *
 *
 */
void getResourceData(zipResourceRequest *zipRRequest) {

	char *resourcePath = zipRRequest->request->path_info;
	char *indexPath;
	int suffixPos;
	unz_file_info zipResourceInfo;

	int pathLen = strlen(resourcePath);

	if( !resourcePath || !*resourcePath ) {
		sendZipInfo(zipRRequest);
		return;
	}

	indexPath = resourcePath+1;

	if(resourcePath[pathLen-1] != '/') {

		if(unzLocateFile(zipRRequest->zipFile,indexPath,CMP_CASE_SENSITIVE) != UNZ_OK) {
			sendErrorMsg(zipRRequest,"Can't locate file in archive");
			return; 
		}

	} else {

		indexPath = apr_pstrcat(zipRRequest->request->pool,resourcePath+1,HTTP_INDEX_FILE1,NULL);

		if(unzLocateFile(zipRRequest->zipFile,indexPath,CMP_CASE_SENSITIVE) != UNZ_OK)
			indexPath = apr_pstrcat(zipRRequest->request->pool,resourcePath+1,HTTP_INDEX_FILE2,NULL);

		if(unzLocateFile(zipRRequest->zipFile,indexPath,CMP_CASE_SENSITIVE) != UNZ_OK) {						
				listZipDirectory(zipRRequest);
				return;
			}
	};

	zipRRequest->currentFilePath = indexPath;

	suffixPos = strlen(indexPath) - 1;
    while ( (suffixPos >= 0) && (indexPath[suffixPos] != '/') ) suffixPos--;

    zipRRequest->currentFileName = apr_pstrdup(zipRRequest->request->pool,zipRRequest->currentFilePath+suffixPos+1);

	if(unzGetCurrentFileInfo(zipRRequest->zipFile,&zipResourceInfo,NULL,0,NULL,0,NULL,0) != UNZ_OK) {
		sendErrorMsg(zipRRequest,"Can't retreive information about current file in archive");
		return;
	}

	zipRRequest->currentFileInfo = zipResourceInfo;

	if( (zipRRequest->flags & SEND_DEFLATED_DATA == SEND_DEFLATED_DATA) && (zipResourceInfo.compression_method == Z_DEFLATED) )
		sendCompResource(zipRRequest); else
		sendUncompResource(zipRRequest);					
}

/* 
 * The return value instructs the caller concerning what happened and what to
 * do next:
 *  OK ("we did our thing")
 *  DECLINED ("this isn't something with which we want to get involved")
 *  HTTP_mumble ("an error status should be reported")
 */
static int zipvfs_handler(request_rec *request) {

	char *filename = request->filename;
	const char *acceptEnc;
	char *encodeType;
	unzFile zipFile;

	zipResourceRequest *zipRRequest;


	// TODO: parse string ? (.zip,.jar), but not neccessary
	// TODO: share open file(s) between requests (zipfile pooling)
	
	zipFile = unzOpen2(filename,&zipFuncs);

	if(!zipFile) return DECLINED;

	// allocate momery for our structure
	zipRRequest = apr_palloc(request->pool, sizeof(zipResourceRequest));

	zipRRequest->request = request;
	zipRRequest->zipFile = zipFile;
	zipRRequest->flags = 0;
	
	// register cleanup handler
	apr_pool_cleanup_register(request->pool,zipFile,zipCleanup,apr_pool_cleanup_null);
	

	if( acceptEnc = apr_table_get(request->headers_in,HTTP_ACCEPT_ENCODING) ) {

		while( *(encodeType = ap_getword(request->pool,&acceptEnc,',')) > '\0') {
			
			apr_collapse_spaces(encodeType,encodeType);

			if(!apr_strnatcasecmp(encodeType,HTTP_ENCODE_GZIP1) ||
				!apr_strnatcasecmp(encodeType,HTTP_ENCODE_GZIP2) )
					zipRRequest->flags |= SEND_DEFLATED_DATA;
		}		
	}

	getResourceData(zipRRequest);

	// unregister cleanup handler
	apr_pool_cleanup_kill(request->pool,zipFile,zipCleanup);

	unzClose(zipRRequest->zipFile);

    return OK;
}


/* 
 * Define Hook functions
 *
 *
 *
 *
 *
 */   
void zipvfs_register_hooks(apr_pool_t *p)
{
	// initialize low-level file handlers
	fill_win32_filefunc(&zipFuncs);

    // Hook for the handlers
    ap_hook_handler(zipvfs_handler, NULL, NULL, APR_HOOK_MIDDLE); 
}


/*
 * Module definition for configuration. 
 *
 * The list of callback routines and data structures that provide
 * the static hooks into module from the other parts of the server. 
 *
 */ 

module AP_MODULE_DECLARE_DATA zipvfs_module;

module zipvfs_module =
{
    STANDARD20_MODULE_STUFF,    // module compatabile only for Apache 2.x
                                // standard major,minor,cookie,... stuff
    NULL,                       // per-directory config creator
    NULL,                       // dir config merger
    NULL,						// server config creator
    NULL,                       // server config merger
    NULL,                       // command table 
    zipvfs_register_hooks       // set up other request processing hooks
};

